"use strict";
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _Webhook_instances, _Webhook_patch, _Webhook_destroy, _Webhook_getThreadTarget;
Object.defineProperty(exports, "__esModule", { value: true });
exports.Webhook = void 0;
// Node Imports
const node_events_1 = __importDefault(require("node:events"));
// Local classes
const Message_1 = require("./Message");
const handler_1 = require("../rest/handler");
// Local constants & types
const constants_1 = require("../util/constants");
const types_message_1 = require("../util/types.message");
class Webhook extends node_events_1.default {
    constructor(options) {
        // Ensure presence of required options
        if (!options.id || !options.token) {
            throw new Error("Webhook ID and token are required");
        }
        super();
        _Webhook_instances.add(this);
        this.id = options.id;
        this.token = options.token;
        this.options = {
            autoFetch: false,
            ratelimiterOffset: 0,
            requestTimeout: 15000,
            ...options,
        };
        this.requestHandler = new handler_1.RequestHandler(this);
        if (this.options.autoFetch) {
            this.fetch();
        }
    }
    static from(target, options) {
        if (!(target instanceof URL)) {
            target = new URL(target);
        }
        if (!options)
            options = {};
        const urlValidation = target.protocol === "https:" &&
            target.hostname === "discord.com" &&
            (target.pathname.startsWith("/api/webhooks/") ||
                target.pathname.startsWith(`${constants_1.BASE_URL}/webhooks`));
        if (!urlValidation) {
            throw new Error("Invalid webhook URL");
        }
        const [token, id] = target.pathname.split("/").reverse();
        return new Webhook({ ...options, id, token });
    }
    async fetch() {
        const data = await this.requestHandler.request("GET", constants_1.ROUTER.api.webhook(this.id, this.token));
        __classPrivateFieldGet(this, _Webhook_instances, "m", _Webhook_patch).call(this, data);
        return this;
    }
    async edit(options) {
        if (!options.name && !options.avatar) {
            throw new Error("No valid options were given");
        }
        const data = await this.requestHandler.request("PATCH", constants_1.ROUTER.api.webhook(this.id, this.token), {
            avatar: options.avatar,
            name: options.name,
        });
        __classPrivateFieldGet(this, _Webhook_instances, "m", _Webhook_patch).call(this, data);
        return this;
    }
    /**
     *
     * @param id The ID of the message to fetch
     * @param threadID The ID of the thread to fetch the message from
     * @returns The message
     *
     * @future Optional message caching (likely breaking change, to include the possibility of not providing `threadID`)
     *
     * Possibility to provide the 2nd argument as an object to allow the bypass to force a request.
     * `{ threadID: string, forceFetch: boolean }`
     */
    async fetchMessage(id, threadID) {
        const route = __classPrivateFieldGet(this, _Webhook_instances, "m", _Webhook_getThreadTarget).call(this, constants_1.ROUTER.api.webhookMessage(this.id, this.token, id), { threadID });
        const data = await this.requestHandler.request("GET", route);
        if (!data)
            throw new Error(`Message '${id}' not found`);
        return new Message_1.Message(this, data);
    }
    /**
     *
     * @param content The content of the message
     * @param options The options for the message
     * @returns The created message
     */
    async sendMessage(content, options) {
        if (typeof content !== "string")
            options = content;
        else if (typeof options !== "object")
            options = {};
        if (typeof options !== "object")
            throw new Error("Invalid options");
        options = { ...options };
        if (!options.content && typeof content === "string")
            options.content = content;
        if (!options.content && !options.embeds && !options.files)
            throw new Error("No content, embeds or files");
        // @ts-expect-error
        if (options.threadID && options.thread_name)
            throw new Error("Cannot create a thread (thread_name) and send to an existing one (threadID).");
        // Failover to prevent sending components on a user-owned webhook
        // Only kicks in if the webhook instance has succeeded in fetching it's data
        if (this.channelID && !this.applicationID && options.components) {
            throw new Error("Cannot send components in a webhook without an application ID");
        }
        if (options.suppressEmbeds && !options.flags)
            options.flags = types_message_1.MessageFlags.SUPPRESS_EMBEDS;
        const route = __classPrivateFieldGet(this, _Webhook_instances, "m", _Webhook_getThreadTarget).call(this, constants_1.ROUTER.api.webhook(this.id, this.token), { threadID: options.threadID });
        const data = await this.requestHandler.request("POST", route, {
            allowed_mentions: options.allowed_mentions,
            avatar_url: options.avatar_url,
            components: options.components,
            content: options.content,
            embeds: options.embeds,
            flags: options.flags,
            tts: options.tts,
            username: options.username,
            // @ts-expect-error
            thread_name: options.thread_name,
        }, options.files);
        return new Message_1.Message(this, data);
    }
    // Async sendSlackMessage(options: any) {
    //	return this.requestHandler.request('POST', ROUTER.webhook(this.id, this.token).slack(), false, options);
    // }
    // async sendGitHubMessage(options: any) {
    //	return this.requestHandler.request('POST', ROUTER.webhook(this.id, this.token).github(), false, options);
    // }
    // option to check cache before sending, or force request
    async editMessage(messageID, content, options) {
        if (typeof content !== "string") {
            options = content;
        }
        else if (typeof options !== "object") {
            options = {};
        }
        if (!options.content &&
            !options.embeds &&
            !options.components &&
            !options.files &&
            !options.attachments) {
            throw new Error("No valid options were given");
        }
        const route = __classPrivateFieldGet(this, _Webhook_instances, "m", _Webhook_getThreadTarget).call(this, constants_1.ROUTER.api.webhookMessage(this.id, this.token, messageID), options);
        const data = await this.requestHandler.request("PATCH", route, {
            attachments: options.attachments,
            allowed_mentions: options.allowed_mentions,
            components: options.components,
            content: options.content,
            embeds: options.embeds,
            flags: options.flags,
        }, options.files);
        return new Message_1.Message(this, data);
    }
    // Option to check cache before sending, or force request
    async deleteMessage(messageID, threadID) {
        const query = new URLSearchParams();
        if (threadID)
            query.set("thread_id", threadID);
        const route = __classPrivateFieldGet(this, _Webhook_instances, "m", _Webhook_getThreadTarget).call(this, constants_1.ROUTER.api.webhookMessage(this.id, this.token, messageID), { threadID });
        await this.requestHandler.request("DELETE", route);
    }
    /**
     * This removes the webhook from the channel, and makes the class instance unusable from this point.
     * @returns void
     */
    async delete() {
        await this.requestHandler.request("DELETE", constants_1.ROUTER.api.webhook(this.id, this.token));
        __classPrivateFieldGet(this, _Webhook_instances, "m", _Webhook_destroy).call(this);
    }
}
exports.Webhook = Webhook;
_Webhook_instances = new WeakSet(), _Webhook_patch = function _Webhook_patch(data) {
    if (data.application_id) {
        this.applicationID = data.application_id;
    }
    if (data.avatar) {
        this.avatar = data.avatar;
    }
    if (data.name) {
        this.name = data.name;
    }
    if (data.channel_id) {
        this.channelID = data.channel_id;
    }
    if (data.guild_id) {
        this.guildID = data.guild_id;
    }
}, _Webhook_destroy = function _Webhook_destroy() {
    this.id = "";
    this.token = "";
}, _Webhook_getThreadTarget = function _Webhook_getThreadTarget(route, options) {
    const query = new URLSearchParams({ wait: "true" });
    if (options.threadID) {
        query.set("thread_id", options.threadID);
    }
    return `${route}?${query}`;
};
